open B8_types
open B8_prop

type flat_elt =
  | Space of prop
  | Text of string * prop

type tag =
  | TB | TEM | TI | TPL | TS | TTT | TU
  | T0 | T1 | T2 | T3 | T4 | T5 | T6 | T7 | T8 | T9
  | Tr | Tg | Tb | Tc | Tm | Ty | Tk | Tw

(* fonctions a la con pour les tag *)

let print_tag t =
  Printf.printf "%d " (Obj.magic t)

let tag_of_color = function
  | Red -> Tr
  | Green -> Tg
  | Blue -> Tb
  | Cyan -> Tc
  | Magenta -> Tm
  | Yellow -> Ty
  | Black -> Tk
  | White -> Tw
  | Unspecified -> assert false

let tag_of_size = function
  | B8_types.T0 -> T0
  | B8_types.T1 -> T1
  | B8_types.T2 -> T2
  | B8_types.T3 -> T3
  | B8_types.T4 -> T4
  | B8_types.T5 -> T5
  | B8_types.T6 -> T6
  | B8_types.T7 -> T7
  | B8_types.T8 -> T8
  | B8_types.T9 -> T9
  | B8_types.Tnone -> assert false

(****************************)
(* creation de la structure *)
(****************************)

(* coupe une string en fonction des espaces du debut et de la fin *)
let cut s =
  let len = String.length s in
  if len == 0
  then false, "", false
  else if len == 1
  then
    if is_space s.[0]
    then true, "", false
    else false, s, false
  else
    if is_space s.[0]
    then
      if is_space s.[len-1]
      then true, String.sub s 1 (len-2), true
      else true, String.sub s 1 (len-1), false
    else
      if is_space s.[len-1]
      then false, String.sub s 0 (len-1), true
      else false, s, false

(* donne un representation a plat *)
let flat_of_tdoc d =
  let rec aux cont d =
    match d.expr with
    | Ttag_B d -> aux cont d
    | Ttag_EM d -> aux cont d
    | Ttag_I d -> aux cont d
    | Ttag_PL d -> aux cont d
    | Ttag_S d -> aux cont d
    | Ttag_TT d -> aux cont d
    | Ttag_U d -> aux cont d
    | Ttag_color (d,c) -> aux cont d
    | Ttag_size (d,t) -> aux cont d
    | Tlist l ->
        let rec flat cont = function
          | [] -> cont
          | a::b -> aux (flat cont b) a in
        flat cont l
    | Tstring s ->
        if d.prop.tt
        then Text (s, d.prop)::cont
        else
          (let deb, mid, fin = cut s in
          let contf = if fin then (Space d.prop)::cont else cont in
          let contm = if mid="" then contf else Text (mid, d.prop)::contf in
          if deb then Space d.prop :: contm else contm) in
  aux [] d

(*****************)
(* optimisations *)
(*****************)

(* teste si 2 prop sont compatible pour des textes *)
let text_compat p1 p2 =
  p1 = p2

(* teste si 2 prop sont compatibles pour des espaces *)
(* pour space, on a toujours tt = false *)
let space_compat p1 p2 =
  p1.u == p2.u
    && p1.size == p2.size
    && (p1.color == p2.color || p1.u == 0)

(* teste si 2 prop sont compat entre text et space *)
(* PAS COMMUTATIF *)
let mix_compat pspc ptxt =
  space_compat pspc ptxt && not ptxt.tt

(* optimise sur le plat : peep hole *)
let opt_on_flat f =
  let rec aux res = function
    | [] -> res
    | Text (t1, p1) :: Text (t2, p2) :: l when text_compat p1 p2 ->
        aux res (Text (t1^t2, p2) :: l)
    | Space p1 :: Space p2 :: l when space_compat p1 p2 ->
        aux res (Space p2 :: l)
    | Space p1 :: Text (t2, p2) :: l when mix_compat p1 p2 ->
        (match res with
        | Text (t0, p0) :: r when text_compat p0 p2 ->
            let nt =
              if t0.[(String.length t0) - 1] == ' '
                || t2.[0] == ' '
              then t0 ^ t2
              else t0 ^ " " ^ t2 in
            aux r (Text (nt, p0) :: l)
        | _ ->
            let nt = if t2.[0] == ' ' then t2 else " " ^ t2 in
            aux res (Text (nt, p2) :: l))
    | Space p1 as s1 :: l ->
        (match res with
        | Text (t2, p2) :: r when mix_compat p1 p2 ->
            let nt =
              if t2.[(String.length t2) - 1] == ' '
              then t2 else t2 ^ " " in
            aux r (Text (nt, p2) :: l)
        | _ ->
            aux (s1 :: res) l)
    | a::b -> aux (a::res) b in
  List.rev (aux [] f)

(******************)
(* reconstruction *)
(******************)

let rec add i tag res =
  if i == 0 then res else add (i-1) tag (tag :: res)

(* tag a inserer entre deux textes *)
let text_diff p1 p2 =
  let ct = ref [] and ot = ref [] in
  let manage tag b1 b2 =
    if b1 != b2 then
      if b1 then ct := tag :: !ct else ot := tag :: !ot in
  manage TB p1.b p2.b;
  manage TEM p1.em p2.em;
  manage TI p1.i p2.i;
  manage TS p1.s p2.s;
  manage TTT p1.tt p2.tt;
  if p1.u > p2.u
  then ct := add (p1.u - p2.u) TU !ct
  else ot := add (p2.u - p1.u) TU !ot;
  if p1.size != p2.size then
    (if p1.size != Tnone then ct := (tag_of_size p1.size) :: !ct;
     if p2.size != Tnone then ot := (tag_of_size p2.size) :: !ot);
  if p1.color != p2.color then
    (if p1.color != Unspecified then ct := (tag_of_color p1.color) :: !ct;
     if p2.color != Unspecified then ot := (tag_of_color p2.color) :: !ot);
  !ct, !ot

(* tag a inserer entre deux espaces *)
let space_diff p1 p2 =
  let ct = ref [] and ot = ref [] in
  if p1.u > p2.u
  then ct := add (p1.u - p2.u) TU !ct
  else ot := add (p2.u - p1.u) TU !ot;
  if p1.size != p2.size then
    (if p1.size != Tnone then ct := (tag_of_size p1.size) :: !ct;
     if p2.size != Tnone then ot := (tag_of_size p2.size) :: !ot);
  if p1.color != p2.color then
    (if p1.color != Unspecified then ct := (tag_of_color p1.color) :: !ct;
     if p2.color != Unspecified then ot := (tag_of_color p2.color) :: !ot);
  !ct, !ot

(* tag a inserer entre un texte et un espace *)
let mix_diff p1 p2 =
  let ct, ot = space_diff p1 p2 in
  if p1.tt != p2.tt then
    if p1.tt then (TTT :: ct), ot else ct, TTT :: ot
  else
    ct, ot

(* tag a inserer entre 2 textes separes par des espace *)
let skip_space_diff p1 p2 =
  let ct = ref [] and ot = ref [] in
  let manage tag b1 b2 =
    if b1 != b2 then
      if b1 then ct := tag :: !ct else ot := tag :: !ot in
  manage TB p1.b p2.b;
  manage TEM p1.em p2.em;
  manage TI p1.i p2.i;
  manage TS p1.s p2.s;
  manage TTT p1.tt p2.tt;
  !ct, !ot

(* compte le nombre d'occurence dans une liste *)
let rec compte t res = function
  | [] -> res
  | a::l -> if a==t then compte t (res+1) l else compte t res l

(* trouve la pos de fin d'un tag *)
let rec pos_fin tag opt clt nb pos =
  let rec dcr nb = function
    | [] -> nb
    | t::l -> if t == tag then dcr (nb-1) l else dcr nb l in
  let nb' = nb - (compte tag 0 clt.(pos)) in
  if nb' <= 0
  then pos
  else pos_fin tag opt clt (compte tag nb' opt.(pos)) (pos+1)

(* forget a tag in a list *)
let rec forget t = function
  | [] -> assert false
  | a::l -> if a==t then l else a :: (forget t l)

let rec tag_at fin deb tags opt clt =
  let efft = List.fold_right forget clt.(deb) tags in
  if deb == fin
  then efft
  else tag_at fin (deb+1) (List.rev_append opt.(deb) efft) opt clt

let make_tree d = function
  | TB -> Dtag_B d
  | TEM -> Dtag_EM d
  | TI -> Dtag_I d
  | TPL -> Dtag_PL d
  | TS -> Dtag_S d
  | TTT -> Dtag_TT d
  | TU -> Dtag_U d
  | (T0 | T1 | T2 | T3 | T4 | T5 | T6 | T7 | T8 | T9) as t ->
      Dtag_color (d, Obj.magic 
                    ((Obj.magic B8_types.T0) - (Obj.magic T0) + (Obj.magic t)))
  | (Tr | Tg | Tb | Tc | Tm | Ty | Tk | Tw) as t ->
      Dtag_size (d, Obj.magic
                   ((Obj.magic B8_types.Red) - (Obj.magic Tr) + (Obj.magic t)))

(* cree un arbre de facon simple *)
let dummy_tree f opt clt =
  let get_long l deb =
    let lgt = ref TB and fin = ref 0 in
    let rec aux = function
      | [] -> !lgt, !fin
      | t::ll ->
          let f = pos_fin t opt clt (compte t 0 l)  (deb+1) in
          if f > !fin then (lgt := t; fin := f);
          aux ll in
    aux l in
  let rec aux_tree deb fin =
    Printf.printf "aux_tree %d %d\n" deb fin; flush stdout;
    if opt.(deb) == []
    then
      (let fst =
        (match f.(deb) with
        | Text (t, _) -> Dstring t
        | _ (* Space _ *) -> Dstring " ") in
      if deb + 1 == fin
      then fst
      else
        (match aux_tree (deb+1) fin with
        | Dlist l -> Dlist (fst :: l)
        | d -> Dlist [fst; d]))
    else
      (let lgt, f = get_long opt.(deb) deb in
      opt.(deb) <- forget lgt opt.(deb);
      clt.(f) <- forget lgt clt.(f);
      let tgs = tag_at fin (deb+1) opt.(deb) opt clt in
      opt.(f) <- List.rev_append tgs opt.(f);
      clt.(f) <- List.rev_append tgs clt.(f);
      let fst = make_tree (aux_tree deb f) lgt in
      if f == fin
      then fst
      else
        (match aux_tree f fin with
        | Dlist l -> Dlist (fst :: l)
        | d -> Dlist [fst; d])) in
  let len = Array.length f in
  aux_tree 0 len

(* recree un arbre *)
let doc_of_flat f =
  let lf = Text("", prop_root) :: f @ [Text("", prop_root)] in
  let len = (List.length lf) in
  let opt = Array.create len [] in
  let clt = Array.create len [] in
  let rec fill i last_text pos_text l =
    for i=0 to i do
      Printf.printf "\n%d : " i;
      List.iter print_tag clt.(i);
      print_string " / ";
      List.iter print_tag opt.(i);
    done;
    match l with
    | Text (_, p1) :: ((Text (_, p2) :: _) as l) ->
        let ct, ot = text_diff p1 p2 in
        opt.(i) <- ot; clt.(i) <- ct;
        fill (i+1) p2 (i+1) l (* p2 et i inutiles *)
    | Space pe :: ((Text (_, p2) :: _) as l) ->
        let ce, oe = skip_space_diff last_text p2 in
        let c2, o2 = mix_diff pe p2 in
        if opt.(pos_text) == []
        then
          (opt.(i+1) <- oe@o2;
           if c2 == []
           then clt.(pos_text) <- clt.(pos_text)@ce
           else
             clt.(i+1) <- ce@c2)
        else
          (opt.(pos_text) <- opt.(pos_text)@oe; opt.(i+1) <- o2;
           clt.(pos_text) <- clt.(pos_text)@ce; clt.(i+1) <- c2);
        fill (i+1) p2 i l (* p2 et i inutiles *)
    | Text (_, p1) :: ((Space p2 :: _) as l) ->
        let ct, ot = mix_diff p1 p2 in
        opt.(i) <- ot; clt.(i) <- ct;
        fill (i+1) p1 i l
    | Space p1 :: ((Space p2 :: _) as l) ->
        let ct, ot = space_diff p1 p2 in
        opt.(i) <- ot; clt.(i) <- ct;
        fill (i+1) last_text pos_text l
    | _ -> (* fin *)
        () in
  fill 0 prop_root 0 lf;
  print_string "end list\n"; flush stdout;
  for i=0 to len-1 do
    Printf.printf "\n%d : " i;
    List.iter print_tag clt.(i);
    print_string " / ";
    List.iter print_tag opt.(i);
  done;
  dummy_tree (Array.of_list f) opt clt

(*****************)
(* main function *)
(*****************)

let print_elt = function
  | Space _ -> print_string "SPACE\n"
  | Text (t, _) -> Printf.printf "TEXT ===%s|||\n" t

(* optimisation globale *)
let flat_opt d =
  print_string "flat1\n"; flush stdout;
  let flat1 = flat_of_tdoc d in
  print_string "flat2\n"; flush stdout;
  let flat2 = opt_on_flat flat1 in
  print_string "doc\n"; flush stdout;
  let doc = doc_of_flat flat2 in
  print_string "tdoc\n"; flush stdout;
  let tdoc = B8_typedoc.typedoc doc in
  print_string "end flat\n"; flush stdout;
  tdoc

(*
let _ =
  register "flat" flat_opt 0
*)
