let argc = Array.length Sys.argv
let time = Unix.gettimeofday ()
let delay = ref (-1)
let dump_doc = ref false
let dump_tdoc = ref false
let use_opts = ref []
let dump_size = ref false
let quiet = ref false
let save_temps = ref false
let dump_opts = ref false
let valid_opts = ref false
let valid_result = ref false
let dump_exn = ref false
  
let do_quiet _ = 
  quiet := true; dump_size := true

let do_S _ =
  do_quiet (); save_temps := true

let tdoc_of_ic ic =
  let lexbuf = Lexing.from_channel ic in
  let doc = B8_parser.main  B8_lexer.token lexbuf in
  let tdoc = B8_typedoc.typedoc doc in
  tdoc

let set_valid _ =
  valid_opts := true;
  valid_result := true

let do_debug _ =
  do_S (); dump_opts := true;
  set_valid ();
  dump_exn := true
  
let do_valid f =
  let ic = open_in f in
  Printf.printf "Reading %s" f; print_newline ();
  let tdoc1 = tdoc_of_ic ic in
  Printf.printf "Reading stdin"; print_newline ();
  let tdoc2 = tdoc_of_ic stdin in
  Printf.printf "Check validity"; print_newline ();
  if B8_valid.validate tdoc1 tdoc2 then
    (Printf.printf "Both files are equivalent"; print_newline (); exit 0)
  else
    (Printf.printf "Files are different"; print_newline (); 
      B8_emit.escape := true;
      B8_emit.emit tdoc1 "nequiv1.sml";
      B8_emit.emit tdoc2 "nequiv2.sml";
      exit 1)
  
let _ =
  Arg.parse [
    "-dump_doc", Arg.Set dump_doc,   " : dump doc after parsing ";
    "-dump_tdoc", Arg.Set dump_tdoc, " : dump tdoc after typedoc ";
    "-dump_size", Arg.Set dump_size, " : print input and output sizes ";
    "-quiet", Arg.Unit do_quiet, " : don't output result + -dump_size";
    "-S", Arg.Unit do_S, " : don't remove temporary files + -quiet";
    "-dump_opts", Arg.Set dump_opts, " : dump opt names";
    "-debug", Arg.Unit do_debug, " : -S -dump_opts";
    "-g", Arg.Unit do_debug, " : -S -dump_opts";
    "-valid", Arg.String do_valid, " <filename>: compare stdin with <filename>";
    "-v", Arg.Set valid_result, " : check validity of result";
    "-vv", Arg.Set valid_opts, " : check validity after each opt";
    "-vvv", Arg.Unit set_valid, " : -v -vv";
  ] (fun s -> 
      if !delay < 0 then 
        delay := int_of_string s
      else
        use_opts := s :: !use_opts) "contest delay [opts]"
  
let use_opts = List.rev !use_opts
let delay = !delay

(* Computes the time left for optimizations. Should be dependent on 
the input size ! *)
let opt_time delay =
  max 1 (delay - 10) (* 10 seconds kept for output *)
  
let output_result tdoc =
(* output result *)
  let filename = Filename.temp_file "contest" ".tmp" in
  B8_emit.emit tdoc filename;
  let st = Unix.stat filename in
  if !dump_size then begin
      Printf.printf "Final size: %d" st.Unix.st_size;
      print_newline ();
    end;
  if not !quiet then begin
      let ic = open_in filename in
      let s = String.create 8192 in
      let rec iter () =
        let n = input ic s 0 8129 in
        if n > 0 then
          (output stdout s 0 n; iter ())
      in
      iter ();
      close_in ic;
    end;
  if not !save_temps then Sys.remove filename else
    (Printf.printf "Result: %s" filename; print_newline ());
  exit 0

let _ =
  let lexbuf = Lexing.from_channel stdin in
  let doc = B8_parser.main  B8_lexer.token lexbuf in
  if !dump_size then begin
      Printf.printf "Initial size: %d" (Lexing.lexeme_end lexbuf);
      print_newline ();
    end;
  if !dump_doc then begin
      Printf.printf "Initial text:"; print_newline ();
      B8_dump.dump_doc doc;
      print_newline ();
    end;
  let tdoc = B8_typedoc.typedoc doc in
  if !dump_tdoc then begin
      Printf.printf "Final text:"; print_newline ();
      B8_dump.dump_tdoc tdoc;
      print_newline ();
    end;
  
  let first_doc = tdoc in
  let tdoc = ref tdoc in
  let tdoc_size = ref (B8_valid.size !tdoc) in
  if !quiet then begin
      Printf.printf "Computed initial size %d" !tdoc_size;
      print_newline ();
    end;
  if delay >= 0 then begin
      ignore (Sys.signal Sys.sigalrm (Sys.Signal_handle (fun _ -> 
              if !quiet then begin
                  Printf.printf "TIMEOUT"; print_newline ();
                end;
              let td = if not !valid_result ||
                  B8_valid.validate first_doc !tdoc then
                  !tdoc
                else first_doc in
              output_result td)));
      ignore (Unix.setitimer  Unix.ITIMER_REAL { 
          Unix.it_value = float_of_int (opt_time delay);
          Unix.it_interval = 0.0 })
    end;
(* optimisations *)
  if use_opts = [] then
    let opts = Sort.list (fun (_,(_,i)) (_,(_,j)) -> i < j) !B8_types.opts in
    List.iter (fun (s,(f,_)) ->
        if !dump_opts then
          (Printf.printf "OPT: %s" s; print_newline ());
        try 
          let td = f !tdoc in
          if !save_temps then begin
              let filename = Filename.temp_file s ".tmp" in
              Printf.printf "OUTPUT %s" filename; print_newline ();
              B8_emit.emit td filename;
            end;
          let new_size = B8_valid.size td in
          if !quiet then begin
              Printf.printf "OPT %s output size: %d" s new_size;
              print_newline ();
            end;
          if new_size > !tdoc_size then
            begin
              if !quiet then begin
                  Printf.printf "OPT %s output bigger document !!!" s;
                  print_newline ();
                end;              
            end
          else
          if not !valid_opts || B8_valid.validate td first_doc then
            tdoc := td 
          else
            if !quiet then begin
              Printf.printf "OPT %s output incorrect result !!!" s;
              print_newline ();
            end
        with e -> 
            if !dump_exn then
              (Printf.printf "OPT %s generate exception %s" s (
                  Printexc.to_string e); print_newline ())
    ) opts
  else
    List.iter (fun s ->        
        let (f,_) = try 
            List.assoc s !B8_types.opts
          with _ ->
              Printf.printf "Unable to find OPT: %s" s;
              print_newline ();
              exit 1
        in
        if !dump_opts then
          (Printf.printf "OPT: %s" s; print_newline ());
        try 
          let td = f !tdoc in
          if !save_temps then begin
              let filename = Filename.temp_file s ".tmp" in
              Printf.printf "OUTPUT %s" filename; print_newline ();
              B8_emit.emit td filename;
            end;
          let new_size = B8_valid.size td in
          if !quiet then begin
              Printf.printf "OPT %s output size: %d" s new_size;
              print_newline ();
            end;
          if new_size > !tdoc_size then
            begin
              if !quiet then begin
                  Printf.printf "OPT %s output bigger document !!!" s;
                  print_newline ();
                end;              
            end
          else
          if not !valid_opts || B8_valid.validate td first_doc then
            tdoc := td
          else
          if !quiet then begin
              Printf.printf "OPT %s output incorrect result !!!" s;
              print_newline ();
            end
        with e -> 
            if !dump_exn then
              (Printf.printf "OPT %s generate exception %s" s (
                  Printexc.to_string e); print_newline ())
    ) use_opts;
  let td = if not !valid_result ||
      B8_valid.validate first_doc !tdoc then
      !tdoc
    else first_doc in
  output_result td
  
